#include "draw.h"

using namespace std;

CDraw::CDraw(const float& coef_resol_) : CGraphic({0, 0}, coef_resol_), texture(NULL), destimg(NULL) {
}

CDraw::CDraw(SDL_Surface *img, const SDL_Rect& coord_, const float& coef_resol_) : texture(NULL), destimg(NULL), CGraphic(img, coord_, coef_resol_) {

   // We get the texutre format RGB or BGR
   texture_format_GL();

   // only one texture
   nbr_textures = 1;
   texture = new GLuint[nbr_textures];

   // Have OpenGL generate a texture object handle for us
   glGenTextures( 1, texture);

   // Bind the texture object
   glBindTexture( GL_TEXTURE_2D, texture[0] );

   // Set the texture's stretching properties
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

   // if width or height are not in power 2
   if(object_properties().w & (object_properties().w - 1) || object_properties().h & (object_properties().h - 1)) {

      // We calculate the near power 2 number
      int w = pow2sup(object_properties().w);
      int h = pow2sup(object_properties().h);

      // We set alpha at 0 to copy all alpha channel
      SDL_SetAlpha(surface, 0, 0);

      // We create a new image with good dimensions bigger than surface
      destimg = SDL_CreateRGBSurface(SDL_HWSURFACE, w, h, 32, surface->format->Rmask, surface->format->Gmask, surface->format->Bmask, surface->format->Amask);

      // We blit surface into the new image with dim in power 2
      SDL_BlitSurface(surface, NULL, destimg, NULL);

      // We point on good image with new properties
      surface = destimg;

      get_object_properties(destimg);

   }

   // We create the texture
   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, object_properties().w, object_properties().h, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, surface->pixels );
}

CDraw::CDraw(const string& file, const SDL_Rect& coord_, const float& coef_resol_) : destimg(NULL), texture(NULL), CGraphic(coord_, coef_resol_) {

   // We load the image
   try {
      surface = IMG_Load(file.c_str());
      if (!surface)
         throw string(string("draw.cpp : Can't load the image ") + file);
   }
   catch(string caught) {
      cerr << caught << endl;
      throw;
   }

   // We can contruct the mother with image
   new (this) CGraphic(surface, coord_, coef_resol_);

   // We get the texutre format RGB or BGR
   texture_format_GL();

   // If width texture is big, we cut it
   nbr_textures = 1;

   if(object_properties().w > SIZE_TEXTURE_W_MAX)
      nbr_textures = object_properties().w / SIZE_TEXTURE_W_MAX;
 
   texture = new GLuint[nbr_textures];

   // for every texture
   for(unsigned i = 0; i < nbr_textures; ++i) {
      // Have OpenGL generate a texture object handle for us
      glGenTextures( 1, texture + i);

      // Bind the texture object
      glBindTexture( GL_TEXTURE_2D, texture[i] );

      // Set the texture's stretching properties
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
      glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

      if(object_properties().w > SIZE_TEXTURE_W_MAX) {

         // The new surface where we copy the zone from surface
         SDL_Surface *sdest = NULL;

         // We set alpha at 0 to copy all alpha channel
         SDL_SetAlpha(surface, 0, 0);

         // We create the new surface with maximal width
         sdest = SDL_CreateRGBSurface(SDL_HWSURFACE, SIZE_TEXTURE_W_MAX, object_properties().h, surface->format->BitsPerPixel, surface->format->Rmask, surface->format->Gmask, surface->format->Bmask, surface->format->Amask);

         SDL_Rect src, dest;
         src.x = SIZE_TEXTURE_W_MAX * i;
         src.y = 0;
         src.w = SIZE_TEXTURE_W_MAX;
         src.h = object_properties().h;
         dest.x = 0;
         dest.y = 0;

         // we blit the zone into the new image
         SDL_BlitSurface(surface, &src, sdest, &dest);

         // we create the texture from image
         glTexImage2D( GL_TEXTURE_2D, 0, sdest->format->BytesPerPixel, sdest->w, sdest->h, 0,
                       texture_format, GL_UNSIGNED_BYTE, sdest->pixels );

         // We free the zone image
         SDL_FreeSurface(sdest);
      }
      else
         // we create the texture from image
         glTexImage2D( GL_TEXTURE_2D, 0, surface->format->BytesPerPixel, object_properties().w, object_properties().h, 0,
                       texture_format, GL_UNSIGNED_BYTE, surface->pixels );
   }
}

void CDraw::Free_surface() {

   if(surface) {
      SDL_FreeSurface(surface);
      surface = NULL;
   }

}

CDraw::~CDraw() {

   if(texture) {
      glDeleteTextures(nbr_textures, texture);
      delete[] texture;
   }

   if(surface)
      SDL_FreeSurface(surface);
}

void CDraw::blit_SDL(const bool& reverse, SDL_Surface *surf, SDL_Rect *src, SDL_Rect dest)
{
      if(!reverse)
         // we blit surface into surf
         SDL_BlitSurface(surface, src, surf, &dest);
      else
         // we blit surf into surface
         SDL_BlitSurface(surf, src, surface, &dest);

   // We get the texutre format RGB or BGR
   texture_format_GL();

   // only one texture
   nbr_textures = 1;
   texture = new GLuint[nbr_textures];

   // Have OpenGL generate a texture object handle for us
   glGenTextures( 1, texture);

   // Bind the texture object
   glBindTexture( GL_TEXTURE_2D, texture[0] );

   // Set the texture's stretching properties
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR );
   glTexParameteri( GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR );

   // if width or height are not in power 2
   if(object_properties().w & (object_properties().w - 1) || object_properties().h & (object_properties().h - 1)) {

      // We calculate the near power 2 number
      int w = pow2sup(object_properties().w);
      int h = pow2sup(object_properties().h);

      // We set alpha at 0 to copy all alpha channel
      SDL_SetAlpha(surface, 0, 0);

      // We create a new image with good dimensions bigger than surface
      destimg = SDL_CreateRGBSurface(SDL_HWSURFACE, w, h, 32, surface->format->Rmask, surface->format->Gmask, surface->format->Bmask, surface->format->Amask);

      // We blit surface into the new image with dim in power 2
      SDL_BlitSurface(surface, NULL, destimg, NULL);

      // We point on good image with new properties
      surface = destimg;

      get_object_properties(destimg);
   }

   // We create the texture
   glTexImage2D( GL_TEXTURE_2D, 0, GL_RGBA, object_properties().w, object_properties().h, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, surface->pixels );

}

SDL_Rect CDraw::get_properties() {
   SDL_Rect surf_rect;
   surf_rect.x = object_properties().x;
   surf_rect.y = object_properties().y;
   surf_rect.w = surface->w;
   surf_rect.h = surface->h;
   return surf_rect;
}

SDL_Color CDraw::get_pixel_color_SDL(const TVector2d<unsigned>& coord) {

   SDL_LockSurface(surface);
   SDL_Color pixel_color;

    /* Here p is the address to the pixel we want to retrieve */
    Uint8 *p = (Uint8 *)surface->pixels + coord.y * surface->pitch + coord.x * surface->format->BytesPerPixel;

   // is it a correct pixel format ? Are we in the good zone with x and y ?
   try {

      if(coord.x > object_properties().w || coord.y > object_properties().h)
         throw string("draw.cpp : x or y are too big to locate the pixel");

      if(surface->format->BytesPerPixel != 4 && surface->format->BytesPerPixel != 3)
         throw string("draw.cpp : Bytes Per Pixel is not good");

   }
   catch(string caught) {
      cerr << caught << endl;
      throw;
   }

   // What is the architecture machine ?
   if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
   {
      pixel_color.r = p[2];
      pixel_color.g = p[1];
      pixel_color.b = p[0];
   }
   else {
      pixel_color.b = p[2];
      pixel_color.g = p[1];
      pixel_color.r = p[0];
   }

   SDL_UnlockSurface(surface);
   return pixel_color;
}

void CDraw::full_color_SDL(const SDL_Color& col) {

   SDL_LockSurface(surface);

   // is it a correct pixel format in 32 bits ?
   try {
      if(surface->format->BytesPerPixel != 4)
         throw string("draw.cpp : Bytes Per Pixel is not 3 sorry");
   }
   catch(string caught) {
      cerr << caught << endl;
      throw;
   }

   // for all pixels
   for(unsigned i = 0; i < object_properties().h * object_properties().w; ++i)

      // What is the architecture machine ?
      if(SDL_BYTEORDER == SDL_BIG_ENDIAN)
         *((Uint32 *)surface->pixels + i) = (col.r << 16) | (col.g << 8) | col.b;
      else
         *((Uint32 *)surface->pixels + i) = (col.b << 16) | (col.g << 8) | col.r;

   SDL_UnlockSurface(surface);

}

void CDraw::reverse_H_SDL() {

   int i;
   Uint8 *pixel1, *pixel2;

   SDL_LockSurface(surface);

   // line per line
   for (int y = 0; y < surface->h; y++) {

      // pixel1 point on the first line pixel
      pixel1 = (Uint8*)surface->pixels + y * surface->pitch;

      // pixel2 point on the last line pixel
      pixel2 = pixel1 + surface->w * surface->format->BytesPerPixel;

      // when we are in midle for this line, it's the end for this line
      while (pixel1 < pixel2) {

         pixel2 -= surface->format->BytesPerPixel;

         // we swap pixe1 and pixel2
         for (i = 0; i < surface->format->BytesPerPixel; ++i) {
            Uint8 tmp = *(pixel2 + i);
            *(pixel2 + i) = *(pixel1 + i);
            *(pixel1 + i) = tmp;
         }

         pixel1 += surface->format->BytesPerPixel;
      }
   }

   SDL_UnlockSurface(surface);

}

void CDraw::show_GL(const bool& orientation, const SDL_Rect& dest, const float& deep) {

   float gl_x = (float)dest.x * coef_resol;
   float gl_y = (float)dest.y * coef_resol;
   float gl_w = (float)dest.w * coef_resol;
   float gl_h = (float)dest.h * coef_resol;

   //  Bind the texture object
   glBindTexture( GL_TEXTURE_2D, texture[0] );
   glPushMatrix();
   glTranslatef(gl_x, SIZE_TEXTURE_H_MAX * coef_resol - gl_y, deep);
   glBegin(GL_QUADS);
   
   // We draw the rect texture
   if(!orientation) {
      glTexCoord2i(0,0);
      glVertex3i(0, 0,-1);
      glTexCoord2i(1,0);
      glVertex3i(gl_w,0,-1);
      glTexCoord2i(1,1);
      glVertex3i(gl_w,-gl_h,-1);
      glTexCoord2i(0,1);
      glVertex3i(0,-gl_h,-1);
   }
   else {
      glTexCoord2i(1,0);
      glVertex3i(0, 0,-1);
      glTexCoord2i(0,0);
      glVertex3i(gl_w,0,-1);
      glTexCoord2i(0,1);
      glVertex3i(gl_w,-gl_h,-1);
      glTexCoord2i(1,1);
      glVertex3i(0,-gl_h,-1);
   }
   
   glEnd();
   glPopMatrix();

}

void CDraw::show_GL(const bool& orientation, const SDL_Rect& dest, const TVector2d<float>& scale, const float& deep) {

   float gl_x = (float)dest.x * coef_resol;
   float gl_y = (float)dest.y * coef_resol;
   float gl_w = (float)dest.w * coef_resol;
   float gl_h = (float)dest.h * coef_resol;

   //  Bind the texture object
   glBindTexture( GL_TEXTURE_2D, texture[0] );
   glPushMatrix();
   glTranslatef(gl_x, SIZE_TEXTURE_H_MAX * coef_resol - gl_y, deep);
   glScalef(scale.x, scale.y, 1);
   glBegin(GL_QUADS);
   
   // We draw the rect texture
   if(!orientation) {
      glTexCoord2i(0,0);
      glVertex3i(0, 0,-1);
      glTexCoord2i(1,0);
      glVertex3i(gl_w,0,-1);
      glTexCoord2i(1,1);
      glVertex3i(gl_w,-gl_h,-1);
      glTexCoord2i(0,1);
      glVertex3i(0,-gl_h,-1);
   }
   else {
      glTexCoord2i(1,0);
      glVertex3i(0, 0,-1);
      glTexCoord2i(0,0);
      glVertex3i(gl_w,0,-1);
      glTexCoord2i(0,1);
      glVertex3i(gl_w,-gl_h,-1);
      glTexCoord2i(1,1);
      glVertex3i(0,-gl_h,-1);
   }
   
   glEnd();
   glPopMatrix();

}

void CDraw::show_text_GL(const bool& orientation, const SDL_Rect& dest, const TVector2d<float>& text_coord, const float& deep) {

   float gl_x = (float)dest.x * coef_resol;
   float gl_y = (float)dest.y * coef_resol;
   float gl_w = (float)dest.w * coef_resol * text_coord.x;
   float gl_h = (float)dest.h * coef_resol * text_coord.y;

   //  Bind the texture object
   glBindTexture( GL_TEXTURE_2D, texture[0] );
   glPushMatrix();
   glTranslatef(gl_x, SIZE_TEXTURE_H_MAX * coef_resol - gl_y, deep);
   glBegin(GL_QUADS);
   
   // We draw the rect texture
   if(!orientation) {
      glTexCoord2f(0,0);
      glVertex3i(0, 0,-1);
      glTexCoord2f(text_coord.x,0);
      glVertex3i(gl_w,0,-1);
      glTexCoord2f(text_coord.x,text_coord.y);
      glVertex3i(gl_w,-gl_h,-1);
      glTexCoord2f(0,text_coord.y);
      glVertex3i(0,-gl_h,-1);
   }
   else {
      glTexCoord2f(text_coord.x,0);
      glVertex3i(0, 0,-1);
      glTexCoord2f(0,0);
      glVertex3i(gl_w,0,-1);
      glTexCoord2f(0,text_coord.y);
      glVertex3i(gl_w,-gl_h,-1);
      glTexCoord2f(text_coord.x,text_coord.y);
      glVertex3i(0,-gl_h,-1);
   }
   
   glEnd();
   glPopMatrix();

}
